from datetime import datetime, timezone
from typing import List,Tuple,Dict,Union
from .open_remote import open_remote_forecast

def extract_along_spotter_tracks():
    pass


def extract_forecast_along_tracks( init_time ):
    pass
    # dataset = open_remote_forecast(
    #     variable=variables,
    # )